/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.facade.rest.interceptor;

import java.io.UnsupportedEncodingException;
import java.net.URLDecoder;
import java.net.URLEncoder;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.Response.Status;
import javax.ws.rs.ext.Provider;

import org.apache.commons.codec.binary.Base64;
import org.jboss.resteasy.annotations.interception.Precedence;
import org.jboss.resteasy.annotations.interception.ServerInterceptor;
import org.jboss.resteasy.core.Headers;
import org.jboss.resteasy.core.ResourceMethod;
import org.jboss.resteasy.core.ServerResponse;
import org.jboss.resteasy.spi.Failure;
import org.jboss.resteasy.spi.HttpRequest;
import org.jboss.resteasy.spi.interception.PreProcessInterceptor;
import org.ow2.bonita.facade.APIAccessor;
import org.ow2.bonita.facade.ManagementAPI;
import org.ow2.bonita.facade.impl.StandardAPIAccessorImpl;
import org.ow2.bonita.identity.auth.APIMethodsSecurity;
import org.ow2.bonita.identity.auth.DomainOwner;
import org.ow2.bonita.util.AccessorUtil;
import org.ow2.bonita.util.BonitaConstants;
import org.ow2.bonita.util.xml.XStreamUtil;

import com.thoughtworks.xstream.XStream;

/**
 * This class do the security for the REST API. All incoming HTTP requests passed here.
 * @author Elias Ricken de Medeiros
 *
 */
@Provider
@ServerInterceptor
@Precedence("SECURITY")
public class LoginPreProcessorInterceptor implements PreProcessInterceptor {

  private static Logger LOG = Logger.getLogger(LoginPreProcessorInterceptor.class.getName());
  private static String AUTHENTICATION_SCHEME = "Basic";
  private static String OPTIONS = "options";	
  private static boolean optionsMapCreated = false;

  public ServerResponse preProcess(HttpRequest request, ResourceMethod method)
  throws Failure, WebApplicationException {    
    //get the options map from form parameter
    List<String> options =  request.getDecodedFormParameters().get(OPTIONS);
    Map<String, String> optionsMap = null;
    if (options != null && !options.isEmpty()) {
      String strOptions;
      try {
        strOptions = URLDecoder.decode(options.get(0), "UTF-8");
      } catch (UnsupportedEncodingException e) {
        strOptions = options.get(0);
      }
      optionsMap = getOptionsMap(strOptions);			
    } else {
      optionsMap = getOptionsMap(null);
    }
    
    if (optionsMap.get(APIAccessor.QUERYLIST_OPTION) == null) {
      optionsMap.put(APIAccessor.QUERYLIST_OPTION, AccessorUtil.QUERYLIST_DEFAULT_KEY);
    }
    if (optionsMap.get(APIAccessor.DOMAIN_OPTION) == null){		
      optionsMap.put(APIAccessor.DOMAIN_OPTION, BonitaConstants.DEFAULT_DOMAIN);
    }
    
    //if it's not a secure method skip authentication
    if (!isRESTServerSecured() || !APIMethodsSecurity.isSecuredMethod(method.getMethod())) {
      //update the options map in the http parameters
    	if (optionsMapCreated){
    		updateOptionsFormParam(request, optionsMap);
    	}
      return null;
    }

    //set domain
    DomainOwner.setDomain(optionsMap.get(APIAccessor.DOMAIN_OPTION));

    String passwordHash = optionsMap.get(APIAccessor.PASSWORD_HASH_OPTION);		
    //password is on the options' map
    if (passwordHash != null) {
      String queryList = optionsMap.get(APIAccessor.QUERYLIST_OPTION);
      final ManagementAPI manegementAPI = new StandardAPIAccessorImpl().getManagementAPI(queryList);
      String username = optionsMap.get(APIAccessor.USER_OPTION);
      boolean allowed = manegementAPI.checkUserCredentialsWithPasswordHash(username, passwordHash);
      if (allowed){
        //authenticated
        return null;
      } else {
        //return unauthorized response
        return getServerUnauthorizedResponse();
      }
    }

    //password is on the header
    HttpHeaders headers = request.getHttpHeaders();
    List<String> authorization = headers.getRequestHeader("Authorization");
    if (authorization == null || authorization.isEmpty()) {			
      return getServerUnauthorizedResponse();
    } 

    //get username and password
    String encodedUserPassword = authorization.get(0).replaceFirst(AUTHENTICATION_SCHEME + " ", "");
    String usernameAndPassword = new String(Base64.decodeBase64(encodedUserPassword.getBytes()));

    StringTokenizer tokenizer = new StringTokenizer(usernameAndPassword, ":");
    String username = tokenizer.nextToken();
    String password = tokenizer.nextToken();

    //fill out the options map
    optionsMap.put(APIAccessor.USER_OPTION, username);

    //verify user en and password
    String queryList = optionsMap.get(APIAccessor.QUERYLIST_OPTION);
    final ManagementAPI manegementAPI = new StandardAPIAccessorImpl().getManagementAPI(queryList);
    boolean	allowed = manegementAPI.checkUserCredentials(username, password);
    if (!allowed){
      return getServerUnauthorizedResponse();
    }		
    //update the options map in the http parameters
    updateOptionsFormParam(request, optionsMap);
    //authenticated
    return null;
  }

  /**
	 * @return
	 */
	private boolean isRESTServerSecured() {
		boolean isSecured = true;
		String strIsSecured = System.getProperty(BonitaConstants.ACTIVATE_REST_AUTHENTICATION_PROPERTY);
		
		if (strIsSecured != null && strIsSecured.equalsIgnoreCase("false")){
			isSecured = false;
		}
		
		return isSecured;
	}

	/**
   * Get the unauthorized response
   * @return
   */
  private ServerResponse getServerUnauthorizedResponse() {
    Headers<Object> metadata = new Headers<Object>();
    metadata.add("WWW-Authenticate", getServerResponseHeaders());
    return new ServerResponse(null, Status.UNAUTHORIZED.getStatusCode(), metadata);
  }

  private void updateOptionsFormParam(HttpRequest request,
      Map<String, String> optionsMap) {
    String options = getStringRepresentation(optionsMap);
    //form parameters
    request.getFormParameters().remove(OPTIONS);
    request.getFormParameters().add(OPTIONS, getStringRepresentation(optionsMap));

    //decoded form parameters
    request.getDecodedFormParameters().remove(OPTIONS);
    request.getDecodedFormParameters().add(OPTIONS, options);
  }

  private String getStringRepresentation(Map<String, String> optionsMap) {
    XStream xstream = XStreamUtil.getDefaultXstream();
    String strRepresntation = xstream.toXML(optionsMap);
    try {
      strRepresntation = URLEncoder.encode(strRepresntation, "UTF-8");
    } catch (UnsupportedEncodingException e) {
      if (LOG.isLoggable(Level.WARNING)) {
        LOG.warning("Cannot encode " + strRepresntation + " using UTF-8");
      }
    }
    return strRepresntation;
  }

  private String getServerResponseHeaders() {		
    String response = AUTHENTICATION_SCHEME + " realm=\"REST_API\"";		
    return response;
  }

  @SuppressWarnings("unchecked")
  private Map<String, String> getOptionsMap(String strOptions) {
    if (strOptions == null || "".equals(strOptions.trim())) {
    	optionsMapCreated = true;
      return new HashMap<String, String>();
    }

    XStream xstream = null;
    //xml representation
    if (strOptions.startsWith("<")){
      xstream = XStreamUtil.getDefaultXstream();
      try {
        return (Map<String,String>) xstream.fromXML(strOptions);
      } catch (Exception e) {
      	optionsMapCreated = true;
        return new HashMap<String, String>();
      }
    }
    
    optionsMapCreated = true;
    Map<String, String> map = new HashMap<String, String>();
    //parse the user entry
    StringTokenizer optionsTokenizer = new StringTokenizer(strOptions, ",");

    while(optionsTokenizer.hasMoreTokens()) {
      String option = optionsTokenizer.nextToken();
      StringTokenizer elementTokenizer = new StringTokenizer(option, ": ");
      if (elementTokenizer.countTokens() == 2){
        String key = elementTokenizer.nextToken();
        String value = elementTokenizer.nextToken();				
        if (key.equalsIgnoreCase(APIAccessor.QUERYLIST_OPTION)) {
          map.put(APIAccessor.QUERYLIST_OPTION, value);
        }
        if (key.equalsIgnoreCase(APIAccessor.DOMAIN_OPTION)) {
          map.put(APIAccessor.DOMAIN_OPTION, value);
        }
        if (key.equalsIgnoreCase(APIAccessor.USER_OPTION)){
        	map.put(APIAccessor.USER_OPTION, value);
        }
      }
    }
    return map;
  }

}
