/**
 * Copyright (C) 2010  BonitaSoft S.A.
 * BonitaSoft, 31 rue Gustave Eiffel - 38000 Grenoble
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.bonita.util;

import java.io.InputStream;
import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.Map.Entry;

import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.SQLQuery;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.cfg.Configuration;
import org.ow2.bonita.env.EnvConstants;
import org.ow2.bonita.env.GlobalEnvironmentFactory;
import org.ow2.bonita.facade.IdentityAPI;
import org.ow2.bonita.facade.def.InternalActivityDefinition;
import org.ow2.bonita.facade.def.element.impl.MetaDataImpl;
import org.ow2.bonita.facade.identity.impl.GroupImpl;
import org.ow2.bonita.facade.identity.impl.MembershipImpl;
import org.ow2.bonita.facade.identity.impl.RoleImpl;
import org.ow2.bonita.facade.identity.impl.UserImpl;
import org.ow2.bonita.facade.privilege.PrivilegePolicy;
import org.ow2.bonita.facade.privilege.Rule.RuleType;
import org.ow2.bonita.facade.privilege.impl.RuleImpl;
import org.ow2.bonita.facade.privilege.impl.RuleTypePolicyImpl;
import org.ow2.bonita.facade.runtime.ActivityState;
import org.ow2.bonita.facade.runtime.TaskInstance;
import org.ow2.bonita.facade.runtime.impl.CategoryImpl;
import org.ow2.bonita.facade.runtime.impl.InternalActivityInstance;
import org.ow2.bonita.facade.runtime.impl.InternalProcessInstance;
import org.ow2.bonita.facade.uuid.ActivityInstanceUUID;
import org.ow2.bonita.light.impl.LightActivityInstanceImpl;
import org.ow2.bonita.runtime.event.EventExecutor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * @author Anthony Birembaut, Matthieu Chaffotte
 */
public final class DbMigration {

  private static final String JAVA_IO_TMPDIR = "java.io.tmpdir";

  private static final Logger LOG = LoggerFactory.getLogger(DbMigration.class);

  private DbMigration() {
  }

  /**
   * Migrate DB schema of the database defined in the environment.
   */
  public static void migrateDb(final String domain, final String configurationName, final String db, final String currentVersion,
      final String targetVersion) throws Exception {
    final SessionFactory sessionFactory = DbTool.getSessionFactory(domain, configurationName.replaceAll("-configuration",
    "-session-factory"));
    Map<String, List<String>> usersRoles = null;
    LOG.info("Running users roles retrieval...");
    final Configuration config = DbTool.getConfiguration(domain, configurationName);
    usersRoles = getUserRoles(config);
    LOG.info("Running " + db + " " + configurationName + " DB migration...");
    final InputStream inputStream = findMigrationScript(db, currentVersion, targetVersion);
    executeScript(sessionFactory, inputStream, db);
    LOG.info("Running users roles migration...");
    setUserRoles(sessionFactory, usersRoles);
    executePostMigration(configurationName, sessionFactory);
    if ("sqlserver".equals(db)) {
      addUniqueConstraints(configurationName, sessionFactory);
    }
    sessionFactory.close();
    LOG.info("---------------Schema migrated---------------");
  }

  private static Map<String, List<String>> getUserRoles(final Configuration config) throws Exception {
    final Map<String, List<String>> usersRoles = new HashMap<String, List<String>>();
    String connectionURL = config.getProperty("hibernate.connection.url");
    final String tmpDirRegex = "${" + JAVA_IO_TMPDIR + "}";
    if (connectionURL.contains(tmpDirRegex)) {
      connectionURL = connectionURL.replace(tmpDirRegex, System.getProperty(JAVA_IO_TMPDIR));
    }
    LOG.info("connecting to db " + connectionURL);
    final Connection conn = DriverManager.getConnection(connectionURL, config.getProperty("hibernate.connection.username"), config.getProperty("hibernate.connection.password"));
    final Statement userStmt = conn.createStatement();
    final ResultSet users = userStmt.executeQuery("select DBID_,USER_UUID_ from BN_USER");
    final PreparedStatement rolesStmt = conn
    .prepareStatement("select ROLE_UUID_ from BN_ROLE ROLE, BN_USER_ROLE USER_ROLE where ROLE.DBID_ = USER_ROLE.ROLE_DBID_ and USER_ROLE.USER_DBID_ = ?");
    while (users.next()) {
      final String userUUID = users.getString("USER_UUID_");
      final long userDbid = users.getLong("DBID_");
      rolesStmt.setLong(1, userDbid);
      final ResultSet roles = rolesStmt.executeQuery();
      final List<String> userRoles = new ArrayList<String>();
      while (roles.next()) {
        userRoles.add(roles.getString("ROLE_UUID_"));
      }
      usersRoles.put(userUUID, userRoles);
    }
    rolesStmt.close();
    userStmt.close();
    conn.close();
    return usersRoles;
  }

  private static void setUserRoles(final SessionFactory sessionFactory, final Map<String, List<String>> usersRoles)
  throws Exception {
    Session session = null;
    try {
      session = sessionFactory.openSession();
      final Transaction tx = session.beginTransaction();

      final GroupImpl defaultGroup = DbTool.createDefaultGroup(session, IdentityAPI.DEFAULT_GROUP_NAME,
          IdentityAPI.DEFAULT_GROUP_LABEL, IdentityAPI.DEFAULT_GROUP_DESCRIPTION, null);

      for (Entry<String, List<String>> userRolesEntry : usersRoles.entrySet()) {
        final Query getUserQuery = session
        .createQuery("SELECT user FROM org.ow2.bonita.facade.identity.impl.UserImpl as user WHERE user.uuid = '"
            + userRolesEntry.getKey() + "'");
        final UserImpl user = (UserImpl) getUserQuery.uniqueResult();
        for (String roleUUID : userRolesEntry.getValue()) {
          final Query getMembershipQuery = session
          .createQuery("SELECT membership FROM org.ow2.bonita.facade.identity.impl.MembershipImpl as membership INNER JOIN FETCH membership.group as group INNER JOIN FETCH membership.role as role WHERE group.uuid = '"
              + defaultGroup.getUUID() + "' AND role.uuid = '" + roleUUID + "'");
          getMembershipQuery.setCacheable(true);
          MembershipImpl membership = (MembershipImpl) getMembershipQuery.uniqueResult();
          if (membership == null) {
            final Query getRoleQuery = session
            .createQuery("SELECT role FROM org.ow2.bonita.facade.identity.impl.RoleImpl as role WHERE role.uuid = '"
                + roleUUID + "'");
            final RoleImpl role = (RoleImpl) getRoleQuery.uniqueResult();
            if (role != null) {
              membership = new MembershipImpl();
              membership.setGroup(defaultGroup);
              membership.setRole(role);
              session.save(membership);
            }
          }
          if (membership != null) {
            user.getMemberships().add(membership);
          }
        }
      }
      tx.commit();
    } catch (HibernateException e) {
      LOG.warn("Unable to add the default users in the DB", e);
    } finally {
      if (session != null) {
        session.close();
      }
    }
  }

  private static void executePostMigration(final String configurationName, final SessionFactory sessionFactory)
  throws Exception {
    udpateActivities(configurationName, sessionFactory);
    udpateInstances(configurationName, sessionFactory);
    updateMetadata(sessionFactory);
    addRuleUUIDs(sessionFactory);
    addUserUUIDs(sessionFactory);
    addLabelUUIDs(sessionFactory);
  }

  @SuppressWarnings("unchecked")
  private static void udpateActivities(final String configurationName, final SessionFactory sessionFactory)
  throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();

    // Update ActivityInstanceUUIDs
    final Query activityDefinitionQuery = session.createQuery("SELECT activityDef FROM org.ow2.bonita.facade.def.InternalActivityDefinition AS activityDef");
    final List<InternalActivityDefinition> activityDefinitions = activityDefinitionQuery.list();
    if (activityDefinitions != null && !activityDefinitions.isEmpty()) {
      for (InternalActivityDefinition activityDefinition : activityDefinitions) {
        String query = "SELECT activity FROM org.ow2.bonita.facade.runtime.impl.InternalActivityInstance AS activity WHERE activity.activityDefinitionUUID.value = '" + activityDefinition.getUUID().getValue() + "'";
        final Query activityQuery = session.createQuery(query);
        final List<InternalActivityInstance> activityInstances = activityQuery.list();
        if (activityInstances != null && !activityInstances.isEmpty()) {
          for (InternalActivityInstance activityInstance : activityInstances) {
            String loopId = "noLoop";
            if (activityDefinition.isInALoop()) {
              loopId = Misc.getUniqueId("lp");
            }
            StringBuilder uuidBuilder = new StringBuilder(activityInstance.getUUID().getValue());
            uuidBuilder.append("--").append(loopId);

            Field field = LightActivityInstanceImpl.class.getDeclaredField("loopId");
            field.setAccessible(true);
            field.set(activityInstance, loopId);
            ActivityInstanceUUID uuid = new ActivityInstanceUUID(uuidBuilder.toString());
            field = LightActivityInstanceImpl.class.getDeclaredField("uuid");
            field.setAccessible(true);
            field.set(activityInstance, uuid);
          }
        }
      }
    }
    session.getTransaction().commit();
    session.close();
  }
  
  @SuppressWarnings("unchecked")
  private static void udpateInstances(final String configurationName, final SessionFactory sessionFactory)
  throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
    LOG.info("Updating process instances ...");
    final Session session = sessionFactory.openSession();
    // Process instance involved/active users migration
    final Query instancesQuery = session.createQuery("SELECT instance FROM org.ow2.bonita.facade.runtime.impl.InternalProcessInstance AS instance");
    final boolean inHistory = EnvConstants.HB_CONFIG_HISTORY.equals(configurationName);
    boolean updateInstance = true;
    int maxResults = 500;
    int currentPosition = 0;
    while (updateInstance) {
      session.getTransaction().begin();
      instancesQuery.setMaxResults(maxResults);
      instancesQuery.setFirstResult(currentPosition);
      final List<Object> processInstances = instancesQuery.list();
      for (Object processInstance : processInstances) {
        final InternalProcessInstance instance = (InternalProcessInstance) processInstance;
        final Set<TaskInstance> tasks = instance.getTasks();
        for (TaskInstance task : tasks) {
          if (task.getState() == ActivityState.READY || task.getState() == ActivityState.EXECUTING) {
            if (task.isTaskAssigned()) {
              instance.addActiveUser(task.getTaskUser());
            } else {
              instance.addActiveUsers(task.getTaskCandidates());
            }
          }
          instance.addInvolvedUsers(task.getTaskCandidates());
        }
        instance.setIsArchived(inHistory);
      }
      if (processInstances.size() < maxResults) {
        updateInstance = false;
      } else {
        currentPosition += maxResults;
      }
      session.getTransaction().commit();
    }
    session.close();
  }

  private static void updateMetadata(final SessionFactory sessionFactory)
  throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
    LOG.info("Updating metadata ...");
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    final Query queryMetadata = session.createQuery("SELECT metadata FROM org.ow2.bonita.facade.def.element.impl.MetaDataImpl AS metadata WHERE key = 'Privilege_default_policy'");
    final MetaDataImpl metadata = (MetaDataImpl) queryMetadata.uniqueResult();
    if (metadata != null) {
      PrivilegePolicy meta = PrivilegePolicy.valueOf(metadata.getValue());
      RuleTypePolicyImpl processStart = new RuleTypePolicyImpl(RuleType.PROCESS_START, meta);
      RuleTypePolicyImpl processRead = new RuleTypePolicyImpl(RuleType.PROCESS_READ, meta);
      session.save(processStart);
      session.save(processRead);
      session.delete(metadata);   
    }
    session.getTransaction().commit();
    session.close();
  }

  @SuppressWarnings("unchecked")
  private static void addRuleUUIDs(final SessionFactory sessionFactory)
  throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
    LOG.info("Updating rules ...");
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    // Generate UUID for each rule
    final Query ruleQuery = session.createQuery("SELECT rule FROM org.ow2.bonita.facade.privilege.impl.RuleImpl AS rule");
    final List<RuleImpl> rules = ruleQuery.list();
    if (rules != null) {
      for (RuleImpl rule : rules) {
        Field field = RuleImpl.class.getDeclaredField("uuid");
        field.setAccessible(true);
        field.set(rule, UUID.randomUUID().toString());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  @SuppressWarnings("unchecked")
  private static void addUserUUIDs(final SessionFactory sessionFactory) {
    LOG.info("Updating users ...");
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    //hash the existing users' passwords
    final Query getUsersQuery = session.createQuery("SELECT DISTINCT user FROM org.ow2.bonita.facade.identity.impl.UserImpl AS user");
    final List<Object> users = getUsersQuery.list();
    for (Object user : users) {
      final UserImpl userImpl = (UserImpl)user;
      if (userImpl.getPassword() != null) {
        userImpl.setPassword(Misc.hash(userImpl.getPassword()));
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  @SuppressWarnings("unchecked")
  private static void addLabelUUIDs(final SessionFactory sessionFactory)
  throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
    LOG.info("Updating labels ...");
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    
    final Query getLabelsQuery = session.createQuery("SELECT label FROM org.ow2.bonita.facade.runtime.impl.LabelImpl AS label");
    final List<Object> labels = getLabelsQuery.list();
    if (labels != null) {
      for (Object label : labels) {
        Field field = CategoryImpl.class.getDeclaredField("uuid");
        field.setAccessible(true);
        field.set(label, UUID.randomUUID().toString());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  private static void addUniqueConstraints(String configurationName, SessionFactory sessionFactory)
  throws Exception {
    LOG.info("Add unique constraints");
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    final SQLQuery categoryUUIDConstraint = session.createSQLQuery("alter table BN_CATEGORY add unique(CATEGORY_UUID_)");
    categoryUUIDConstraint.executeUpdate();
    final SQLQuery ruleUUIDConstraint = session.createSQLQuery("alter table BN_RULE add unique(RULE_UUID_);");
    ruleUUIDConstraint.executeUpdate();
    session.getTransaction().commit();
    session.close();
  }

  public static void executeScript(SessionFactory sessionFactory, InputStream inputStream, String db) {
    final byte[] bytes = IoUtil.readBytes(inputStream);
    final String scriptContent = new String(bytes);
    final List<String> commands = getCommands(scriptContent, db);

    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();

    LOG.info("DB Commands Execution");
    for (String command : commands) {
      LOG.info("Executing command : " + command);
      try {
        session.createSQLQuery(command).executeUpdate();
      } catch (Exception e) {
        System.err.println("Error while executing command: " + command);
        LOG.error(e.getMessage(), e.getCause());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  public static List<String> getCommands(final String scriptContent, final String db) {
    String delimiter = ";";
    if (db.equals("sqlserver") || db.equals("sybase")) {
      delimiter = "go";
    }
    String regex = delimiter.concat("\n");
    final List<String> commands = new ArrayList<String>();
    String[] tmp= scriptContent.split(regex);
    for (String command : tmp) {
      commands.add(command.trim());
    }
    int lastIndex = commands.size() - 1;
    String lastCommand = commands.get(lastIndex);
    int index = lastCommand.lastIndexOf(delimiter);
    if (index > 0) {
      lastCommand = lastCommand.substring(0, index);
      commands.remove(lastIndex);
      commands.add(lastCommand);
    }
    return commands;
  }

  public static InputStream findMigrationScript(final String db, final String currentVersion, final String targetVersion) {
    final StringBuilder migrationScript = new StringBuilder();
    migrationScript.append("/migration/").append(db).append("-").append(currentVersion).append("-").append(
        targetVersion).append(".sql");
    LOG.info("Loading Script " + migrationScript.toString());
    final InputStream inputStream = DbMigration.class.getResourceAsStream(migrationScript.toString());
    if (inputStream == null) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_2");
      throw new IllegalArgumentException(message);
    }
    return inputStream;
  }

  private static void removeHistoryDefaultUsers(final String domain) throws Exception {
    final SessionFactory sessionFactory = DbTool.getSessionFactory(domain, EnvConstants.HB_CONFIG_HISTORY.replaceAll("-configuration", "-session-factory"));
    final Session session = sessionFactory.openSession();
    session.getTransaction().begin();
    LOG.info("Remove default users");

    String[] defaultUsers = new String[] {"admin", "john", "jack", "james"};
    for (String user : defaultUsers) {
      String statement = "SELECT DBID_ FROM BN_USER WHERE USER_USERNAME_ = '" + user + "'";
      try {
        Object dbid = session.createSQLQuery(statement).uniqueResult();
        statement = "DELETE FROM BN_USER_ROLE WHERE USER_DBID_ = " + dbid;
        session.createSQLQuery(statement).executeUpdate();
        statement = "DELETE FROM BN_USER WHERE DBID_ = " + dbid;
        session.createSQLQuery(statement).executeUpdate();
      } catch (HibernateException e) {
        System.err.println("Error while executing command: " + statement);
        e.printStackTrace();
        LOG.error(e.getMessage(), e.getCause());
      }
    }
    session.getTransaction().commit();
    session.close();
  }

  public static void main(final String[] args) throws Exception {
    if (args == null || args.length != 4) {
      final String message = ExceptionManager.getInstance().getFullMessage("bh_DBM_1");
      throw new IllegalArgumentException(message);
    }
    final String domain = args[0];
    final String db = args[1].toLowerCase();
    final String currentVersion = args[2];
    final String targetVersion = args[3];
    String disableSearch = "bonita.search.use";
    System.setProperty(disableSearch, "false");
    final EventExecutor eventExecutor = GlobalEnvironmentFactory.getEnvironmentFactory(BonitaConstants.DEFAULT_DOMAIN).get(EventExecutor.class);
    eventExecutor.stop();
    LOG.info("Starting History DB migration");
    removeHistoryDefaultUsers(domain);
    migrateDb(domain, EnvConstants.HB_CONFIG_HISTORY, db, currentVersion, targetVersion);
    LOG.info("Starting Core DB migration");
    migrateDb(domain, EnvConstants.HB_CONFIG_CORE, db, currentVersion, targetVersion);
    System.clearProperty(disableSearch);
  }

}
